import kcs_util
import kcs_ui
import kcs_draft
import kcs_dex

import KcsPoint2D
import KcsModel

import string
import math
from Tkinter import *
import tkMessageBox

def _set(entry, text):
  entry.delete(0, END)
  entry.insert(END, text)

class CabinData:
  """General class handling the file interface for cabin data"""
  SEPARATOR = ";"
  itemPrefix=["Name:", "Module:", "Deck:", "MFZ:", "Room ID:", "Designation:",
              "Assignment:", "Class code:", "Type:", "STG:"]

  def __init__(self, defaultCB_State = None):
    """Fetch location of external files, get project name,
    and establish default state of item checkboxes.
    
    defaultCB_State should be a list of items, whose truth value
    (1 or 0) defines the default state of subsequent checkboxes.
    If None, all checkboxes are set."""
    try:
      self.DB_DIR = kcs_util.TB_environment_get("CA_ACC_DB")
    except: #if not defined, defaults to SB_SHIP
      self.DB_DIR = kcs_util.TB_environment_get("SB_SHIP")
    try:
      self.ROOM_FILE = kcs_util.TB_environment_get("CA_ACC_ROOMPROP")
    except:
      self.ROOM_FILE = "roomprop.dat"

    #Read MFZ data
    try:
      fileName = kcs_util.TB_environment_get("CA_ACC_MFZ")
    except:   
      fileName = "mfz.dat"
    self.MFZ = self.GetModuleClasses(fileName)
    if self.MFZ is None:
      self.DECK = None
      return

    #Read DECK data  
    try:
      fileName = kcs_util.TB_environment_get("CA_ACC_DECK")
    except:
      fileName = "deck.dat"
    self.DECK = self.GetModuleClasses(fileName)

    #get project name
    self.project = kcs_util.TB_environment_get("SB_PROJ_STRUC")

    #Set up default checkbox state list
    nItems = len(self.itemPrefix)
    if defaultCB_State is None:
      self.defCB = [1]*nItems
    else:
      defCB = []
      for n in range(nItems):
        try:
          if defaultCB_State[n]: value = 1
          else: value = 0
        except:
          value = 1
        defCB.append(value)
      self.defCB = defCB

  def error(self):
    """True, if some of the internal structures could not be
    initialised properly."""
    return (self.MFZ is None) or (self.DECK is None)

  def GetModuleClasses(self, FileName):
    """Loads the contents of the given file into memory. The lines must have
    the layout:
    
    className;item1;item2; ... ;itemN;
    
    'Items' must be unique, so that the dictionary of pairs item:className
    can be built.
    The dictionary is returned, or None, if an error occurs."""
    dict = {}
    try:
      f = open(self.DB_DIR + FileName, "r")
      try:
        line = f.readline()
        while line != "":
          items = map(string.strip, string.split(line, self.SEPARATOR))
          if len(items) > 1:
            className = items[0]
            for module in filter(None, items[1:]):
              if dict.has_key(module):
                tkMessageBox.showwarning("Cabin property editor", \
                  "WARNING: Module %s occurs more than once in the file %s!" % (module, FileName))
                return None
              else:
                dict[module] = className
          line = f.readline()
      finally:
        f.close()
    except:
      tkMessageBox.showerror("Cabin property editor", "Error reading the file %s!" % FileName)
      return None
    return dict

  def getRoomProperties(self, name):
    """getRoomData(name) -> (status, [room ID, designation, assignment, Class code, type, STG])
    
    Querries the room data file for the information about the given
    cabin. Status can be set as follows:
       1 - record found, iformation retrieved
       0 - record not found, blank information retrieved
      -1 - error reading the room data file."""
    try:
      f = open(self.DB_DIR + self.ROOM_FILE, "r")
      try:
        line = f.readline()
        SEPARATOR = self.SEPARATOR
        split = string.split
        strip = string.strip
        roomProps = len(self.itemPrefix) - 4
        
        while line != "":
          #Fetch the room name, and its attributes
          props = split(line, SEPARATOR, 1)
          if name == strip(props[0]): #Is it the right room?
            #Isolate room attributes
            items = map(strip, split(props[1], SEPARATOR))
            n = len(items)
            if n <= roomProps:
              while n < roomProps:
                items.append("") #Pad with empty attributes, if not given
                n += 1
              return (1, items)
            else:
              return (1, items[:roomProps])
          line = f.readline()
        return (0, [""]*roomProps)
      finally:
        f.close()
    except:
      return (-1, None)

  def getFullRoomDataset(self, name, module):
    """Return the cabin information suitable for setting the state and
    contents of the window widgets."""
    try:
      deck = self.DECK[module]
    except:
      deck = "not defined"
    try:
      mfz = self.MFZ[module]
    except:
      mfz = "not defined"
    status, data = self.getRoomProperties(name)
    if status >= 0:
      data[0:0] = [name, module, deck, mfz]
      res = zip(data, self.defCB) #entry contents and checkbox state
      return (status, res)
    else:
      return (status, None) #Error fetching room properties
    
  def updateRoomFile(self, name, updatedData):
    """Reads the current contents of the room properties file,
    updates the current cabin information and writes the room information
    back.
    
    Returns 0 on success, 1 on error"""
    #Fetch current contents of the data file
    try:
      f = open(self.DB_DIR + self.ROOM_FILE, "r")
      try:
        lines = f.readlines()
      finally:
        f.close()
      #Temporarily remove trailing blank line
      try:
        if lines[-1] == "\n":
          del lines[-1]
      except:
        pass
    except:
      tkMessageBox.showerror("Cabin property editor", "Error reading the room data file: %s!" % self.ROOM_FILE)
      return 1

    #Open the file for writing, prepare updated data, and write them out
    try:
      f = open(self.DB_DIR + self.ROOM_FILE, "w")
      SEPARATOR = self.SEPARATOR
      rowNo = found = 0
      find = string.find
      strip = string.strip

      for dataRow in lines:
        ind = find(dataRow, SEPARATOR)
        if (ind > 0) and (strip(dataRow[:ind]) == name):
          found = 1
          break
        rowNo += 1

      if found:
        lines[rowNo] = updatedData
      else:
        lines.append(updatedData)
      #Append blank line
      lines.append("\n")

      try:
        f.writelines(lines)
      finally:
        f.close()
      return 0
    except:
      return 1

  def getCabin(self):
    """Picks a structure model object from the drawing. Returns a tuple
    containing the structure name and module, or None, if the selection
    has been cancelled by the user."""
    point = KcsPoint2D.Point2D()
    OK = kcs_util.ok()
    YES = kcs_util.yes()
    while 1:
      res = kcs_ui.point2D_req("Indicate cabin", point)
      if res[0] != OK: return None

      model = KcsModel.Model('struct') #A point has been indicated
      try:    #Try to recognise the model object
        res = kcs_draft.model_identify(point, model)
      except: #Failed
        kcs_ui.message_noconfirm("Cabin not found!")
      else:   #Succeeded, structure object selected!
        st = "STRUCTURE('%s').ITEM('%s').MODULE.NAME" % (self.project, model.Name)
        if (kcs_dex.extract(st) == 0) and (kcs_dex.next_result() == 3):
          module = kcs_dex.get_string() #Get module name
          handle = kcs_draft.element_highlight(res[1])
          #Get user approval
          res = kcs_ui.answer_req("Indicate cabin", \
            "Cabin %s in module %s, OK?" % (model.Name, module))
          kcs_draft.highlight_off(handle)
          if res == YES:
            return (model.Name, module)
        else:
          kcs_ui.message_noconfirm("Cannot get cabin module name!")

#-----------------------------------------------------------------------------------------------
  
class CabinEditor:
  OK = kcs_util.ok()  
  msgWARN = "WARN"
  msgINFO = "INFO"
  msgERR = "ERROR"
  
  def __init__(self, master, cabData):
    """Creates the window contents inside the given master window"""
    self.master = master
    master.title("Cabin property editor")
    self.cabData = cabData

    frame = Frame(master, relief=RAISED, borderwidth=1)
    frame.grid(row=0, column=0, sticky=N, pady=2)
    Label(frame, text="Property", font="Helvetica 10 bold").grid(row=0, column=0)
    Label(frame, text="Value", font="Helvetica 10 bold").grid(row=0, column=1)

    self.entries = []
    self.disabledEntries = [1, 2, 3, 4]
    lab = cabData.itemPrefix
    nItems = len(lab)
    
    for n in range(1, nItems+1):
      e = Entry(frame)
      e.fIndex = n-1
      if n in self.disabledEntries:
        e.configure(bg="Grey")
        e.bind("<Key>", self.ignoreKey)
      e.grid(row=n, column=1, padx=5)
      self.entries.append([e, None])

    defCB = cabData.defCB
    for n in range(nItems):
      v = IntVar()
      v.set(defCB[n])
      b = Checkbutton(frame, variable=v, text=lab[n], command=self.toggle)
      b.grid(row=n+1, column=0, sticky=W)
      item = self.entries[n]
      item[1] = v

    fr1 = Frame(master)
    fr1.grid(row=0, column=1, sticky=S, pady=2)
    Button(fr1, text="Select all", command=self.selAll).grid(row=0, column=0, sticky=W+E, padx=2, pady=3)
    Button(fr1, text="Deselect all", command=self.deselAll).grid(row=1, column=0, sticky=W+E, padx=2, pady=3)
    Button(fr1, text="Update", command=self.updateRoomFile).grid(row=2, column=0, sticky=W+E, padx=2, pady=3)
    self.addB = Button(fr1, text="Add to drawing\n(current item)", command=self.addToDwgCurrent, state=DISABLED)
    self.addB.grid(row=3, column=0, sticky=W+E, padx=2, pady=3)
    self.addB2 = Button(fr1, text="Add to drawing\n(selected items)", command=self.addToDwgSelected)
    self.addB2.grid(row=4, column=0, sticky=W+E, padx=2, pady=3)
    Frame(fr1, height=33).grid(row=5, column=0, sticky=W+E, padx=2)
    Button(fr1, text="OK", command=self.ok).grid(row=6, column=0, sticky=W+E, padx=2, pady=3)
    Button(fr1, text="Close", command=self.close).grid(row=7, column=0, sticky=W+E, padx=2, pady=3)

    self.status = Label(master, text="", relief=SUNKEN, bd=1, justify=LEFT)
    self.status.grid(row=1, column=0, columnspan=2, sticky=N+S+E+W)
    master.bind("<FocusIn>", self.identField)
    self.curField = 4
    self.entries[self.curField][0].focus_set()

  def updateRoomFile(self):
    """Reads the current contents of the room properties file,
    updates the current cabin information and writes the room information
    back.
    
    Returns 0 if all went OK, otherwise some non-zero status"""
    curCabin = self.entries[0][0].get()
    updatedData, updatedDefs = curCabin, []
    SEPARATOR = self.cabData.SEPARATOR
    for n in range(4, len(self.cabData.itemPrefix)):
      item = self.entries[n][0].get()
      updatedData = updatedData + SEPARATOR + item
      updatedDefs.append(item)
      
    updatedData = updatedData + SEPARATOR + "\n"
    res = self.cabData.updateRoomFile(curCabin, updatedData)
    if res == 0:
      self.initCabinData = updatedDefs
      self.setStatus("Cabin properties updated successfully!", self.msgINFO)
    else:
      self.setStatus("Error updating room data file!", self.msgERR)
    return res

  def setStatus(self, msg, msgClass):
    """Updates the contents of the status line"""
    if msgClass == self.msgINFO:
      self.status.configure(text=msg, fg="Black", font="Helvetica 8")
    elif msgClass == self.msgWARN:
      self.status.configure(text=msg, fg="Blue", font="Helvetica 8 bold")
    else:
      self.status.configure(text=msg, fg="Red", font="Helvetica 8 bold")

  def initFields(self, data):
    """Prepares the window by filling the cabin data
    and setting up checkbox states."""
    n = 0
    for e, cb in self.entries:
      e.insert(END, data[n][0])
      cb.set(data[n][1])
      n += 1
    self.toggle()
    self.entries[4][0].focus_set()
    self.initCabinData = map(lambda x: x[0], data[4:])
      
  def identField(self, event):
    """Identifies currently active Entry, and sets the 'Add to drawing'
    button state appropriately"""
    try:
      self.curField = event.widget.fIndex
      self.addB.config(state=NORMAL)
    except:
      self.curField = -1
      self.addB.config(state=DISABLED)

  def ignoreKey(self, event):
    """Ignores all key events except the Tab key. This effectively protects
    the entry contents from being modified, but allows tabbing through
    available entries, and getting keyboard focus."""
    if event.char != "\t": return "break"

  def _setAll(self, value):
    """Sets the state of all checkboxes to 'value'"""
    for e in self.entries:
      e[1].set(value)
    self.toggle()
    
  def selAll(self):
    """Turns on all checkboxes"""
    self._setAll(1)

  def deselAll(self):
    """Turns off all checkboxes"""
    self._setAll(0)

  def toggle(self):
    """Determines, how many checkboxes are set, and sets the state
    of the 'Add to drawing' button appropriately."""
    nON = reduce(lambda x, y: x + y[1].get(), self.entries, 0)
    if nON > 0: bstate = NORMAL
    else: bstate = DISABLED
    self.addB2.config(state=bstate)

  def addToDwgCurrent(self):
    """Adds to drawing the currently focused information item"""
    ind = self.curField
    if ind != -1:
      self.addToDwg([self.entries[ind][0].get()])

  def addToDwgSelected(self):
    """Adds to drawing all currently selected information items"""
    list = []
    for item in self.entries:
      if item[1].get():
        list.append(item[0].get())
    self.addToDwg(list)

  def addToDwg(self, msgList):
    """Performs the actual actions related to adding the texts
    in the msgList list to the drawing"""
    #Check and update room props file
    needsUpdate = 0
    for n in range(4, len(self.entries)):
      if self.entries[n][0].get() != self.initCabinData[n]:
        needsUpdate = 1
        break
    if needsUpdate:
      if self.updateRoomFile() != 0:
        return
      
    self.master.withdraw()
    height = kcs_draft.text_height_get()
    wFac = height*kcs_draft.text_aspect_get()*0.5
    rotation = kcs_draft.text_rotation_get()*math.pi/180.0
    hFac = kcs_draft.text_ilsp_get()*height
    sin, cos = math.sin(rotation), math.cos(rotation)
    dX, dY = hFac*sin, -hFac*cos
    point = KcsPoint2D.Point2D()
    res = kcs_ui.point2D_req("Indicate text position", point)
    if res[0] == self.OK:
      x0, y0 = point.X, point.Y
      for msg in msgList:
        n = len(msg)
        if n > 0:
          w2 = n*wFac
          point.SetCoordinates(x0-w2*cos, y0-w2*sin)
          kcs_draft.text_new(msg, point)
        x0 += dX
        y0 += dY
      self.ok()
    else:
      self.master.deiconify()

  def close(self):
    """Closes the window"""
    self.loop = 0 #do not restart, just quit
    self.master.quit()

  def ok(self):
    """Closes the window"""
    #Update default checkboxes state list
    defCB = []
    for item in self.entries:
      defCB.append(item[1].get())
    self.cabData.defCB = defCB
    #Quit the window
    self.loop = 1 #restart, pick a new cabin
    self.master.quit()

def editCabinData(cabinGeneralData, savedData):
  root = Tk()
  try:
    app = CabinEditor(root, cabinGeneralData)
    app.initFields(savedData)
    root.mainloop()
    loop = app.loop
  finally:
    root.destroy()
    del root
  return loop

def getGeneralData(defaultCB_State = None):
  root = Tk()
  try:
    root.withdraw()
    cabinGeneralData = CabinData(defaultCB_State)
  finally:
    root.destroy()
    del root
  return cabinGeneralData
  
def editCabinProperties(defaultCB_State = None):
  cabinGeneralData = getGeneralData(defaultCB_State)
  if cabinGeneralData.error(): return

  status = 1
  while status:
    res = cabinGeneralData.getCabin()
    if res is None: status = 0
    else:
      stat, data = cabinGeneralData.getFullRoomDataset(res[0], res[1])
      if stat < 0: status = 0
      else: status = editCabinData(cabinGeneralData, data)
    
def run():
  editCabinProperties()
  
if __name__ == "__main__":
## use either this option (controling checkboxes state):
  defaultCB_State = [0,0,0,1,1,1,1,0,0,0]
  editCabinProperties(defaultCB_State)

## or this (all checkboxes set)  
#  run()

  